--[[
    https://us.forums.blizzard.com/en/wow/t/wow-classic-ui-api-change-for-unithealth/446596
    https://www.townlong-yak.com/framexml/37176/GlobalStrings.lua/RU
    https://github.com/tomrus88/BlizzardInterfaceCode/blob/master/Interface/SharedXML/SharedColorConstants.lua#L72-L81

    Multiplication factors prefix symbol:
    1e+24 1,000,000,000,000,000,000,000,000 yotta Y
    1e+21 1,000,000,000,000,000,000,000 zetta Z
    1e+18 1,000,000,000,000,000,000 exa E
    1e+15 1,000,000,000,000,000 peta P
    1e+12 1,000,000,000,000 tera T
    1e+09 1,000,000,000 giga G
    1e+06 1,000,000 mega M
    1e+03 1,000 kilo k
    1e+02 100 hecto h
    1e+01 10 deka da
    1e+00 1
    1e-01 0.1 deci d
    1e-02 0.01 centi c
    1e-03 0.001 milli m
    1e-06 0.000 001 micro
    1e-09 0.000 000 001 nano n
    1e-12 0.000 000 000 001 pico p
    1e-15 0.000 000 000 000 001 femto f
    1e-18 0.000 000 000 000 000 001 atto a
    1e-21 0.000 000 000 000 000 000 001 zepto z
    1e-24 0.000 000 000 000 000 000 000 001 yocto y

    Convert RGB(%) to Hex:
    string.format('|cff%02x%02x%02x%s|r', r * 255, g * 255, b * 255, text)

    Issue #1 (fixed):
    Hold LMB or RMB but do not rotate the camera, just move your character
    The cursor should not disappear from the screen at this time, it will follow the movement of your character
    Thus, while it follows, when hovering over a unit there is a color issue
    Because unit returns nil under the given conditions

    Issue #2 (fixed):
    Hover over a unit (during combat) and then move the cursor away
    The tooltip should not disappear from the screen at this time (FadeOut animation)
    Then quickly let the unit take some damage to see the changes
    The status bar color will turn to default green, not class color
    Because unit returns nil under the given conditions

    Issue #3 (fixed):
    An issue similar to issue #2 also occurs when using UnitPlayerControlled()
    Turn off nameplates to see it
]]

local _, addonTable = ...
local project = addonTable.project
local config = addonTable.config
local data = {} -- stores dynamic data

local function abbreviateLargeNumbers(value)
    if value >= 1e11 then return string.format('%dG', value / 1e9) -- 128G
    elseif value >= 1e9 then return string.format('%.1fG', value / 1e9) -- 1.6G, 61.8G
    elseif value >= 1e8 then return string.format('%dM', value / 1e6) -- 128M
    elseif value >= 1e6 then return string.format('%.1fM', value / 1e6) -- 1.6M, 61.8M
    elseif value >= 1e5 then return string.format('%dk', value / 1e3) -- 128k
    elseif value >= 1e3 then return string.format('%.1fk', value / 1e3) -- 1.6k, 61.8k
    else return value
    end
end

local function constrain(value, low, high) -- not isRetail
    return math.max(math.min(value, high), low)
end

local function insertLine(tooltip, position, text, color) -- not isRetail
    -- Text and color do not matter at this step
    tooltip:AddLine(text, 1, 1, 1)
    tooltip:Show() -- update borders correctly

    local numLines = tooltip:NumLines()
    local position = constrain(position, 2, numLines)
    local color = color or CreateColor(1, 1, 1)

    for i = numLines, position, -1 do
        local currLine = _G[tooltip:GetName() .. 'TextLeft' .. i]
        local nextLine = _G[tooltip:GetName() .. 'TextLeft' .. i - 1]

        if i == position then
            currLine:SetText(text)
            currLine:SetTextColor(color.r, color.g, color.b)
            return -- stop the loop and exit the function
        else
            currLine:SetText( nextLine:GetText() )
            currLine:SetTextColor( nextLine:GetTextColor() ) -- because RESURRECTABLE is green
        end
    end
end

-- Fix issue #2 described in the comments above
local function unitClassColor(unit)
    if UnitIsPlayer(unit) then
        data.prevColor = data.currColor
        data.currColor = RAID_CLASS_COLORS[ select(2, UnitClass(unit)) ]
    else
        data.prevColor = data.currColor
        data.currColor = CreateColor(0, 1, 0) -- if not a player or nil
    end

    if unit == nil then
        data.currColor = data.prevColor
    end
    return data.currColor
end

-- Fix issue #3 described in the comments above
local function unitPlayerControlled(unit) -- isClassic_TBC
    data.prevUnit = data.currUnit
    data.currUnit = UnitPlayerControlled(unit)

    if unit == nil then
        data.currUnit = data.prevUnit
    end
    return data.currUnit
end

local function updateStatusBar(statusBar, unit)
    local isDead = unit and UnitIsDeadOrGhost(unit)
    local isPlayerControlled = unitPlayerControlled(unit)
    local unitColor = config.enableClassColor and unitClassColor(unit) or CreateColor(0, 1, 0)
    local currHealth = statusBar:GetValue()
    local maxHealth = select(2, statusBar:GetMinMaxValues())
    local healthText = ''

    -- ChatFrame1:Clear()
    -- print( 'isDead = ' .. tostring(isDead) )
    -- print( 'isPlayerControlled = ' .. tostring(isPlayerControlled) )
    -- print( 'unitColor = ' .. string.format('%.1f, %.1f, %.1f', unitColor.r, unitColor.g, unitColor.b) )

    if not config.showUnitHealth then
        healthText = ''
    elseif currHealth <= 0 or isDead then
        healthText = DEAD
    else
        if project.isClassic_TBC and isPlayerControlled then
            -- 1.13.3 hotfix from 23.02.2020
            -- Now, the UI API functions UnitHealth and UnitHealthMax return health values for NPCs (rather than percentages)
            -- This hotfix has no impact on PVP, as enemy players or the units they summon still return health percentages, as before
            healthText = math.ceil(currHealth/maxHealth * 100) .. '%' -- default rounding for status bar
        else
            healthText = abbreviateLargeNumbers(currHealth) .. ' / ' .. abbreviateLargeNumbers(maxHealth)
        end
    end
    statusBar.healthText:SetText(healthText)
    statusBar:SetStatusBarColor(unitColor.r, unitColor.g, unitColor.b)
end

local function onTooltipSetUnit(self)
    if IsMouseButtonDown('LeftButton') or IsMouseButtonDown('RightButton') then
        return self:Hide() -- fix issue #1 described in the comments above
    end

    local unit = select(2, self:GetUnit())
    local statusBar = self:GetChildren()
    data.currUnit = UnitPlayerControlled(unit) -- isClassic_TBC

    -- ChatFrame1:Clear()
    -- print( 'unitType = ' .. tostring(unit and UnitGUID(unit)):lower():match('%l+') )

    if UnitIsPlayer(unit) then
        local unitName = GetUnitName(unit, config.showServerName) -- also hides a title
        local guildName, _, _, guildRealm = GetGuildInfo(unit) -- isClassic
        local classColor = RAID_CLASS_COLORS[ select(2, UnitClass(unit)) ]
        local guildColor = config.enableGuildColor and ChatTypeInfo['GUILD'] or CreateColor(1, 1, 1)
        local playerFaction = select(2, UnitFactionGroup('player'))
        local unitFaction = select(2, UnitFactionGroup(unit))

        -- Add missing lines in earlier versions of the game
        if not project.isRetail then
            if project.isClassic and guildName then
                insertLine(self, 2, guildRealm and (guildName .. '-' .. guildRealm) or guildName) -- guild line
            end
            insertLine(self, guildName and 4 or 3, unitFaction) -- faction line
        end

        for i = 1, self:NumLines() do
            local line = _G[self:GetName() .. 'TextLeft' .. i]
            local lineText = line:GetText()

            -- Set class color
            if i == 1 then
                line:SetText(
                    -- Fix UnitFrame name color by using classColor.colorStr
                    config.enableClassColor and ('|c' .. classColor.colorStr .. unitName .. '|r') or unitName
                )

            -- Set guild line color
            elseif i == 2 and guildName then
                line:SetText(config.showServerName and lineText or guildName)
                line:SetTextColor(guildColor.r, guildColor.g, guildColor.b)

            -- Set faction line color, if the unit (player) is from the opposite faction
            elseif not config.enableFactionColor then
                break
            elseif (lineText == FACTION_ALLIANCE or lineText == FACTION_HORDE) and playerFaction ~= unitFaction then
                line:SetTextColor(
                    FACTION_BAR_COLORS[1].r,
                    FACTION_BAR_COLORS[1].g,
                    FACTION_BAR_COLORS[1].b
                )
                break
            end
        end
        data.currColor = classColor
    else
        data.currColor = CreateColor(0, 1, 0) -- if not a player or nil
    end
    updateStatusBar(statusBar, unit)
end

local function onValueChanged(self)
    local unit = select(2, self:GetParent():GetUnit())

    updateStatusBar(self, unit)
end

-- In the classic version, the shaman class color was the same as that of the paladin
-- In the TBC expansion, this color was changed from pink to blue
if project.isClassic then
    RAID_CLASS_COLORS['SHAMAN'] = CreateColor(0, 0.44, 0.87) -- blue instead of pink
    RAID_CLASS_COLORS['SHAMAN'].colorStr = RAID_CLASS_COLORS['SHAMAN']:GenerateHexColor()
end

-- Almost useless option, but someone might find it useful
GameTooltip:SetScale(config.scaleFactor)

-- Create a string object for status bar text
GameTooltipStatusBar.healthText = GameTooltipStatusBar:CreateFontString(nil, 'OVERLAY')
GameTooltipStatusBar.healthText:SetFont(config.healthFont, config.healthFontSize, config.healthFontFlags)
GameTooltipStatusBar.healthText:SetPoint('CENTER', 0, config.healthOffsetY)

GameTooltip:HookScript('OnTooltipSetUnit', onTooltipSetUnit)
GameTooltipStatusBar:HookScript('OnValueChanged', onValueChanged)